"""Process git repo URLs."""
import re
from pathlib import Path


def get_repo_url(url: str) -> str:
    """Get repo url by searching for reg like ``https://*/*/*/``.

    Args:
        url: URL path to the external content
    
    Returns:
        repository URL
    """
    pattern = r"https://[^/]+/[^/]+/[^/]+(?=/)"
    match = re.search(pattern, url)
    if match is None:
        msg = (
        "Invalid external content URL. Could not parse repo URL from:\n"
        f" '{url}'"
        )
        raise ValueError(msg)
    return match[0]


def get_branch_tag_name(url: str) -> str:
    """Get branch_tag_name by searching for between blob and book in external_url.

    Args:
        url: URL path to the external content.
    
    Returns:
        branch or tag name.
    """
    pattern = r"blob/([^/]+)/"
    match = re.search(pattern, url)

    if match is None:
        msg = (
        "Invalid external content URL. Could not retrieve branch/tag name from:\n"
        f" '{url}'"
        )
        raise ValueError(msg)
    return match[1]


def create_repository_dir_name(url: str, root_dir: str | Path) -> str:
    """Generate the path where the repo will be cloned to.

    It will be of the form:

    {root_path}/{platform}_{organization}_{repository}/{revision}

    Args:
        url: URL path to the external content.
        root_dir: root directory where the repo will be cloned into.
    
    Returns:
        Path where the repo will be cloned to.
    """
    branch_tag_name = get_branch_tag_name(url)
    url = get_repo_url(url)
    if url.startswith("https://"):
        url = url.removeprefix("https://")
    dir_name = url.replace("/", "_")
    return f"{root_dir}/{dir_name}/{branch_tag_name}"




# from pybtex.style.formatting.unsrt import Style
from pybtex.plugin import register_plugin
from pybtexapastyle.formatting.apa import APAStyle
from pybtexapastyle.labels.apa import LabelStyle as APALabelStyle

# from pybtex.style.template import names, sentence

class MyAPALabelStyle(APALabelStyle):
    def format_label(self, entry):
        return APALabelStyle.format_label(self, entry)

class MyAPAStyle(APAStyle):
    default_label_style = 'myapa'

def setup(app):
    register_plugin('pybtex.style.labels', 'myapa', MyAPALabelStyle)
    register_plugin('pybtex.style.formatting', 'myapastyle', MyAPAStyle)




from pybtex.plugin import Plugin


class BaseSortingStyle(Plugin):
    def sorting_key(self, entry):
        raise NotImplementedError

    def sort(self, entries):
        return sorted(entries, key=self.sorting_key)


"""Utilities for external content."""
from collections.abc import Callable
from pathlib import Path
from typing import Any, overload

import yaml


@overload
def modify_field(
    data: dict, key: str, func: Callable, *args, **kwargs
) -> dict: ...


@overload
def modify_field(
    data: list, key: str, func: Callable, *args, **kwargs
) -> list: ...


def modify_field(
    data: dict | list, key: str, func: Callable, *args, **kwargs
) -> dict | list:
    """Modify the fields that match a given key.

    Recursively look for the fields matching a given key in a YAML-like
    mapping. Modify the matching fields by running `func` on them.

    Args:
        data: mapping where to look for matches
        key: key to look for
        func: function to run on the matching fields
        args: positional arguments for `func`
        kwargs: keyword arguments for `func`
    
    Returns:
        modified mapping
    """
    if isinstance(data, dict):
        if key in data:
            return func(data, *args, **kwargs)
        else:
            return {
                k: modify_field(v, key, func, *args, **kwargs)
                for k, v in data.items()
            }
    elif isinstance(data, list):
        return [modify_field(el, key, func, *args, **kwargs) for el in data]
    return data


def load_yaml_file(
        path: str | Path, encoding: str = "utf8"
) -> dict[str, Any]:
    """Load a yaml file file (ToC or config) as dictionary.

    Args:
        path: file path
        encoding: file character encoding
    
    Returns:
        parsed file
    """
    with open(path, encoding=encoding) as handle:
        data = yaml.safe_load(handle)
    return data



from pathlib import Path

import pytest

from teachbooks.external_content.bib import merge_bibs, read_bibfile, write_bibfile
from tests.test_external_content import BOOK_ROOT, CLONED_REPOS

TEST_BIBFILE = (
    Path(__file__).parent / "testbook" / "book" / "_git" / 
    "gitlab.tudelft.nl_interactivetextbooks-citg_risk-and-reliability" / "main" /
    "book" / "_bibliography" / "references.bib"
)


@pytest.fixture
def merged_bibs():
    return merge_bibs(BOOK_ROOT / "references.bib", CLONED_REPOS)


def test_bib_merge(merged_bibs):
    """Ensure test data reference files are merged correctly."""
    assert len(merged_bibs) == 7 # total unique citekeys

    expected_entries = [
        "baecher2003", # duplicate
        "adk2022", # from risk-and-reliability
        "usace14", # from HOS-workbook
        "jason_moore",  # from excited workshop and main book
    ]
    citekeys = [bib.citekey for bib in merged_bibs]
    for entry in expected_entries:
        assert entry in citekeys


def test_bib_merge_duplicate(capsys, merged_bibs):
    """Warning should be raised when a duplicate entry is encountered."""
    err = capsys.readouterr()
    assert "duplicate citekey 'baecher2003'" in err.err


def test_read_write_bibfile(tmp_path):
    """Make sure that after writing the original bib file has the same content."""
    bibout = tmp_path / "references.bib"

    bibs = read_bibfile(TEST_BIBFILE)

    write_bibfile(bibout, bibs)

    assert bibs == read_bibfile(bibout)




from pybtex.style.sorting import BaseSortingStyle


class SortingStyle(BaseSortingStyle):

    def sorting_key(self, entry):
        if entry.type in ('book', 'inbook'):
            author_key = self.author_editor_key(entry)
        elif 'author' in entry.persons:
            author_key = self.persons_key(entry.persons['author'])
        else:
            author_key = ''
        return (author_key, entry.fields.get('year', ''), entry.fields.get('title', ''))

    def persons_key(self, persons):
        return '   '.join(self.person_key(person) for person in persons)

    def person_key(self, person):
        if person.prelast_names[0] != '{' and person.last_names[0] != '{' and person.first_names[0] != '{':
            return '  '.join((
                ' '.join(person.prelast_names + person.last_names),
                ' '.join(person.first_names + person.middle_names),
                ' '.join(person.lineage_names),
            )).lower()
        else:
            return '  '.join((
                ' '.join(person.first_names + person.middle_names),
                ' '.join(person.prelast_names + person.last_names),
                ' '.join(person.lineage_names),
            )).lower()

    def author_editor_key(self, entry):
        if entry.persons.get('author'):
            return self.persons_key(entry.persons['author'])
        elif entry.persons.get('editor'):
            return self.persons_key(entry.persons['editor'])
        else:
            return ''



from dataclasses import dataclass, field

import sphinxcontrib.bibtex.plugin
from sphinxcontrib.bibtex.style.referencing import BracketStyle
from sphinxcontrib.bibtex.style.referencing.author_year import AuthorYearReferenceStyle


def bracket_style() -> BracketStyle:
    return BracketStyle(
        left='(',
        right=')',
    )


@dataclass
class MyReferenceStyle(AuthorYearReferenceStyle):
    bracket_parenthetical: BracketStyle = field(default_factory=bracket_style)
    bracket_textual: BracketStyle = field(default_factory=bracket_style)
    bracket_author: BracketStyle = field(default_factory=bracket_style)
    bracket_label: BracketStyle = field(default_factory=bracket_style)
    bracket_year: BracketStyle = field(default_factory=bracket_style)


def setup(app):
    sphinxcontrib.bibtex.plugin.register_plugin(
        'sphinxcontrib.bibtex.style.referencing',
        'author_year_round', MyReferenceStyle)


"""Read and process table of contents files."""
import os.path
import stat
import subprocess
from pathlib import Path
from typing import Any

import click
import yaml

from teachbooks.external_content import GIT_PATH
from teachbooks.external_content.bib import merge_bibs, write_bibfile
from teachbooks.external_content.config import check_plugins, get_teachbooks_config
from teachbooks.external_content.git import (
    create_repository_dir_name,
    get_branch_tag_name,
    get_repo_url,
)
from teachbooks.external_content.headers import add_origin_notes
from teachbooks.external_content.licenses import validate_licenses
from teachbooks.external_content.requirements import check_requirements
from teachbooks.external_content.utils import load_yaml_file, modify_field

LOCAL_TOC_HEADER = (
    "ToC file with localized paths. Used for Teachbooks' external content feature."
)


def process_external_toc_entries(
    src: Path,
    dest: Path,
    book_root: Path,
    error_invalid_license: bool = True
) -> Path:
    """Parse external (git) ToC entries, checkout repos & write new ToC to file.
    
    Args:
        src: Path to the source table-of-contents yaml file.
        dest: Path to the destination table-of-contents yaml file.
        book_root: Path to the root of the book.
        error_invalid_license: If true, and no valid license is found in an
            external repository, an error will be raised. Else only a warning.

    Returns:
        Path to the new table-of-contents yaml file, or the original file
            if the toc was not modified.
    """
    src_toc = load_yaml_file(src)
    toc = src_toc.copy()

    toc = modify_field(
        toc,
        key="external",
        func=external_to_local,
        external_path=book_root / GIT_PATH,
        root=book_root,
    )

    cloned_repo_log = book_root / GIT_PATH / "cloned_repos.txt"
    if cloned_repo_log.exists():
        cloned_repos = read_cloned_repos(cloned_repo_log)
        validate_licenses(
            cloned_repos, book_root / GIT_PATH, error_invalid_license
        )
        check_requirements(book_root.parent / "requirements.txt", cloned_repos)
        check_plugins(book_root / "_config.yml", cloned_repos)

        merged_bibs = merge_bibs(book_root / "references.bib", cloned_repos)
        write_bibfile(dest.parent / "references.bib", merged_bibs)
        # TODO: don't overwrite original references.bib file

        write_toc_yaml(toc, dest, header=LOCAL_TOC_HEADER)
        return dest
    
    if toc != src_toc:
        msg = (
            "Table of contents has external git content, "
            "but no git repositories were checked out.\n"
            "Please fix your table of content or remove any `external` entries"
        )
        raise ValueError(msg)
    return src


def read_cloned_repos(log: Path) -> list[Path]:
    """Read in cloned repo file to get paths to all cloned repos."""
    with log.open("r") as f:
        cloned_repos_str = [repo.strip("\n\r") for repo in f.readlines()]
    return [
        Path(repo) if Path(repo).is_absolute() else 
        log.parent / repo for repo in cloned_repos_str
    ]


def get_content_path(url: str) -> str:
    """Get relative path of the external content from the URL path.

    Args:
        url: URL path to the external content
    
    Returns:
        repo path to the external content
    """
    branch_tag_name = get_branch_tag_name(url)
    *_, path = url.split(branch_tag_name)
    return path.strip("/")  # remove leading and trailing "/"


def write_toc_yaml(
    data: dict[str, str],
    path: str | Path,
    encoding: str = "utf8",
    header: str | None = None,
) -> None:
    """Write a ToC file.

    Args:
        data: site map
        path: Target `_toc.yml` file path
        encoding: `_toc.yml` file character encoding
        header: Commented out header to start toc file with.
    """
    with open(path, encoding=encoding, mode="w") as handle:
        if header is not None:
            handle.write(f"# {header}\n")
        yaml.safe_dump(data, handle)


def external_to_local(
    mapping: dict[str, Any],
    external_path: str | Path,
    root: str | Path
) -> dict[str, Any]:
    """Modify mapping with the "external" key.

    Retrieve external components locally, and fix ToC fields accordingly.

    Args:
        mapping: map to modify.
        external_path: path where to store external components.
        root: express paths to external components with respect to root.
    
    Returns:
        map with fields adjusted in order to refer to local resources.
    """
    cfg = get_teachbooks_config(root / "_config.yml")

    mapping_local = mapping.copy()
    external_url = mapping_local.pop("external")

    repo_url = get_repo_url(external_url)
    clone_url = f"{repo_url}.git"

    branch_tag_name = get_branch_tag_name(external_url)
    repository_dir = create_repository_dir_name(external_url, root_dir=external_path)

    cloned_repo_file = Path(external_path) / "cloned_repos.txt"
    if cloned_repo_file.exists():
        with open(cloned_repo_file) as f:
            cloned_repos = f.read()
    else:
        cloned_repos = []

    if repository_dir in cloned_repos:
        click.secho(f"{repository_dir} has already been cloned. Not re-downloading")
    else:
        # clone with branch_name. Check for environment var (for testing)
        if "NO_GIT_CLONE" not in os.environ:
            subprocess.run([
                "git", "clone", "--single-branch", "-b",  branch_tag_name, clone_url,
                repository_dir
            ])
        with cloned_repo_file.open("a") as f:
            f.write(str(repository_dir) + "\n")

        add_origin_notes(Path(repository_dir), cfg, clone_url, version=branch_tag_name)

    content_file = get_content_path(external_url)
    rel_path = os.path.relpath(repository_dir, root)
    mapping_local["file"] = os.path.join(rel_path, content_file).replace("\\", "/")
    return mapping_local


def chmod_git_files(foo, file, err):
    """Remove git files on Windows.

    Solution from: https://stackoverflow.com/a/76356125
    """
    if (
        os.name == "nt" and
        Path(file).suffix in [".idx", ".pack", ".rev"] and
        err[0].__name__ == "PermissionError"
    ):
        os.chmod(file, stat.S_IWRITE)
        foo(file)



"""Parse requirements.txt files."""
from pathlib import Path

import click


def check_requirements(main_requirements: Path, git_repos: list[Path]):
    """Check for mismatch between main and external python dependencies."""
    if not main_requirements.exists():
        click.secho(
            "Warning: no requirements.txt file found for the main book.\n"
            "    TeachBooks will not be able to find which external-content"
            "dependencies are missing",
            fg="yellow",
            err=True,
        )
        return None

    requirements = read_requirements(main_requirements)

    repos = (repo for repo in git_repos if (repo / "requirements.txt").exists())
    for repo in repos:
        repo_requirements = read_requirements(repo / "requirements.txt")
        unmatched_requirements = repo_requirements - requirements
        if len(unmatched_requirements) > 0:
            unmatched_str = "".join(
                [f"       {req}\n" for req in unmatched_requirements]
            )
            click.secho(
                "Warning: requirements do not match between external content "
                "and your book\n"
                f"    Repository: {repo}\n"
                "    Missing or non-matching requirements:\n"
                f"{unmatched_str}"
                "Content from this book might not display or function correctly.",
                fg="yellow",
                err=True,
            )


def read_requirements(file: Path) -> set[str]:
    """Read requirements file and strip comments, index urls, and empty lines."""
    with file.open("r") as f:
        requirements = f.readlines()
    
    valid_requirements = set()
    for req in requirements:
        req = req.strip("\n\r")

        if len(req) == 0 or req.startswith("#") or req.startswith("--"):
            pass
        else:
            valid_requirements.add(req.split(" #")[0])

    return valid_requirements
